__all__ = ["AllowValue"]

from numbers import Number
from typing import Any, NoReturn

from jaxtyping import Array
from plum import dispatch

from . import api
from .base import AbstractQuantity


class AllowValue:
    """A flag to allow a value to be passed through `unxt.ustrip`.

    Examples
    --------
    >>> import jax.numpy as jnp
    >>> import unxt as u
    >>> from unxt.quantity import AllowValue

    >>> x = jnp.array(1)
    >>> y = u.ustrip(AllowValue, "km", x)
    >>> y is x
    True

    >>> u.ustrip(AllowValue, "km", u.Quantity(1000, "m"))
    Array(1., dtype=float32, ...)

    This is a flag, so it cannot be instantiated.

    >>> try:
    ...     AllowValue()
    ... except TypeError as e:
    ...     print(e)
    Cannot instantiate AllowValue

    """

    def __new__(cls) -> NoReturn:
        msg = "Cannot instantiate AllowValue"
        raise TypeError(msg)


@dispatch
def ustrip(flag: type[AllowValue], unit: Any, x: Array | Number, /) -> Array | Number:
    """Strip the units from a value. This is a no-op.

    Examples
    --------
    >>> import jax.numpy as jnp
    >>> import unxt as u
    >>> from unxt.quantity import AllowValue

    >>> x = jnp.array(1)
    >>> y = u.ustrip(AllowValue, "km", x)
    >>> y is x
    True

    """
    return x


@dispatch  # TODO: type annotate by value
def ustrip(flag: type[AllowValue], unit: Any, x: AbstractQuantity, /) -> Any:
    """Strip the units from a quantity.

    Examples
    --------
    >>> import unxt as u
    >>> from unxt.quantity import AllowValue
    >>> q = u.Quantity(1000, "m")
    >>> u.ustrip(AllowValue, "km", q)
    Array(1., dtype=float32, ...)

    """
    return api.ustrip(unit, x)



"""Unitsystem compatibility."""
# TODO: move to a compatibility module

__all__: list[str] = []

from typing import Any

from astropy.units import UnitBase as AstropyUnit
from gala.units import (  # pylint: disable=import-error
    DimensionlessUnitSystem as GalaDimensionlessUnitSystem,
    UnitSystem as GalaUnitSystem,
)
from plum import conversion_method, convert, dispatch

from unxt.unitsystems import AbstractUnitSystem, DimensionlessUnitSystem, dimensionless


@dispatch
def unitsystem(value: GalaUnitSystem, /) -> AbstractUnitSystem:
    """Return a `gala.units.UnitSystem` as a `unxt.AbstractUnitSystem`.

    Examples
    --------
    >>> import unxt as u
    >>> import gala.units as gu
    >>> import astropy.units as apyu

    >>> usys = gu.UnitSystem(apyu.km, apyu.s, apyu.Msun, apyu.radian)

    >>> u.unitsystem(usys)
    unitsystem(km, s, solMass, rad)

    """
    # Create a new unit system instance, and possibly class.
    return unitsystem(*value._core_units)  # noqa: SLF001


@dispatch
def unitsystem(_: GalaDimensionlessUnitSystem, /) -> DimensionlessUnitSystem:
    """Return a `gala.units.DimensionlessUnitSystem` as a `unxt.DimensionlessUnitSystem`.

    Examples
    --------
    >>> import unxt as u
    >>> from gala.units import DimensionlessUnitSystem

    >>> usys = DimensionlessUnitSystem()

    >>> u.unitsystem(usys)
    DimensionlessUnitSystem()

    """  # noqa: E501
    return dimensionless


# =============================================================================
# Convert


@conversion_method(type_from=GalaUnitSystem, type_to=AbstractUnitSystem)  # type: ignore[arg-type]
def convert_gala_unitsystem_to_unxt_unitsystem(
    usys: GalaUnitSystem, /
) -> AbstractUnitSystem:
    """Convert a `gala.units.UnitSystem` to a `unxt.AbstractUnitSystem`.

    Examples
    --------
    >>> import unxt as u
    >>> import gala.units as gu
    >>> import astropy.units as apyu

    >>> usys = gu.UnitSystem(apyu.km, apyu.s, apyu.Msun, apyu.radian)
    >>> usys
    <UnitSystem (km, s, solMass, rad)>

    >>> convert(usys, u.AbstractUnitSystem)
    unitsystem(km, s, solMass, rad)

    """
    return unitsystem(usys)


@conversion_method(type_from=AbstractUnitSystem, type_to=GalaUnitSystem)  # type: ignore[arg-type]
def convert_unxt_unitsystem_to_gala_unitsystem(
    usys: AbstractUnitSystem, /
) -> GalaUnitSystem:
    """Convert a `unxt.AbstractUnitSystem` to a `gala.units.UnitSystem`.

    Examples
    --------
    >>> import unxt as u
    >>> import gala.units as gu
    >>> import astropy.units as apyu
    >>> from plum import convert

    >>> usys = u.unitsystem(apyu.km, apyu.s, apyu.Msun, apyu.radian)
    >>> usys
    unitsystem(km, s, solMass, rad)

    >>> convert(usys, gu.UnitSystem)
    <UnitSystem (km, s, solMass, rad)>

    """

    def convert_unit_to_apyu(u: Any) -> AstropyUnit:
        return convert(u, AstropyUnit)

    return GalaUnitSystem(*map(convert_unit_to_apyu, usys.base_units))



"""Working with dimensions.

The main features are:

- ``unxt.dims.dimension``: a function to construct a dimension object.
- ``unxt.dims.dimension_of``: a function to get the dimensions of an object.

"""

__all__ = ["dimension", "dimension_of", "AbstractDimension"]

from jaxtyping import install_import_hook

from .setup_package import RUNTIME_TYPECHECKER

with install_import_hook("unxt.dims", RUNTIME_TYPECHECKER):
    from ._src.dimensions import AbstractDimension, dimension, dimension_of

# Clean up the namespace
del install_import_hook



"""Configuration file for the Sphinx documentation builder.

This file only contains a selection of the most common options. For a full
list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

from datetime import datetime
from typing import Any

import pytz

from unxt import __version__

# -- Project information -----------------------------------------------------

author = "Unxt Developers"
project = "unxt"
copyright = f"{datetime.now(pytz.timezone('UTC')).year}, {author}"
version = __version__

master_doc = "index"
language = "en"

# -- General configuration ---------------------------------------------------

extensions = [
    "myst_parser",  # General MyST markdown support
    "sphinx_design",
    "sphinx.ext.autodoc",  # TODO: replace with autodoc2
    "sphinx.ext.autosummary",  # TODO: replace with autodoc2
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinx_copybutton",
    "sphinx-prompt",
    "sphinxext.opengraph",
    # "sphinxext.rediraffe",  # Add redirects
    "sphinx_togglebutton",
    "sphinx_tippy",
]

python_use_unqualified_type_names = True

exclude_patterns = [
    "_build",
    "**.ipynb_checkpoints",
    "Thumbs.db",
    ".DS_Store",
    ".env",
    ".venv",
]

source_suffix = [".md", ".rst"]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "jax": ("https://jax.readthedocs.io/en/latest/", None),
    "jaxtyping": ("https://docs.kidger.site/jaxtyping/", None),
    "astropy": ("https://docs.astropy.org/en/stable/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "quax": ("https://docs.kidger.site/quax/", None),
}

# -- Autodoc settings ---------------------------------------------------

autodoc_typehints = "description"
autodoc_typehints_format = "short"

autodoc_default_options = {
    "members": True,
    "undoc-members": True,
    "inherited-members": True,
    "show-inheritance": True,
    "member-order": "bysource",
}

always_document_param_types = True
typehints_use_signature = True


nitpick_ignore = [
    ("py:class", "_io.StringIO"),
    ("py:class", "_io.BytesIO"),
    # TODO: fix these
    ("py:class", "P"),  # ParamSpec alias
    ("py:class", "unxt._src.experimental.R"),  # TypeVar alias
    ("py:class", "unxt._src.quantity.base._QuantityIndexUpdateHelper"),
    ("py:class", "unxt._src.quantity.mixins.AstropyQuantityCompatMixin"),
    ("py:class", "unxt._src.quantity.mixins.IPythonReprMixin"),
    ("py:class", "unxt._src.utils.SingletonMixin"),
    ("py:class", "ArrayLike"),
    ("py:class", "NoneType"),
    ("py:class", "quax._core.ArrayValue"),
    ("py:class", "PhysicalType"),
    ("py:class", "jaxtyping.Shaped[Array, '*shape']"),
    ("py:class", "astropy.units.core.Annotated"),
]

# -- MyST Setting -------------------------------------------------

myst_enable_extensions = [
    "amsmath",  # for direct LaTeX math
    "attrs_block",  # enable parsing of block attributes
    "attrs_inline",  # apply syntax highlighting to inline code
    "colon_fence",
    "deflist",
    "dollarmath",  # for $, $$
    # "linkify",  # identify “bare” web URLs and add hyperlinks:
    "smartquotes",  # convert straight quotes to curly quotes
    "substitution",  # substitution definitions
]
myst_heading_anchors = 3

# myst_substitutions = {
#     "ArrayLike": ":obj:`jaxtyping.ArrayLike`",
#     "Any": ":obj:`typing.Any`",
# }


# -- HTML output -------------------------------------------------

html_theme = "sphinx_book_theme"
html_title = "unxt"
html_logo = "_static/favicon.png"  # TODO: an svg
html_copy_source = True
html_favicon = "_static/favicon.png"

html_static_path = ["_static"]
html_css_files = ["custom_toc.css", "custom_tooltip.css"]

html_theme_options: dict[str, Any] = {
    "home_page_in_toc": True,
    "repository_url": "https://github.com/GalacticDynamics/unxt",
    "repository_branch": "main",
    "path_to_docs": "docs",
    "use_repository_button": True,
    "use_edit_page_button": False,
    "use_issues_button": True,
    "show_toc_level": 2,
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/GalacticDynamics/unxt",
            "icon": "fa-brands fa-github",
        },
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/unxt/",
            "icon": "https://img.shields.io/pypi/v/unxt",
            "type": "url",
        },
        {
            "name": "Zenodo",
            "url": "https://doi.org/10.5281/zenodo.10850455",
            "icon": "fa fa-quote-right",
        },
    ],
}



# pylint: disable=import-error, too-many-lines

"""Test the Array API."""

import pickle

import astropy.units as apyu
import equinox as eqx
import jax
import jax.numpy as jax_xp
import numpy as np
import pytest
from hypothesis import example, given, strategies as st
from hypothesis.extra.array_api import make_strategies_namespace
from hypothesis.extra.numpy import array_shapes as np_array_shapes, arrays as np_arrays
from jax.dtypes import canonicalize_dtype
from jaxtyping import Array, TypeCheckError
from plum import convert, parametric

import quaxed.numpy as jnp

import unxt as u
from unxt.quantity import AbstractParametricQuantity

xps = make_strategies_namespace(jax_xp)


jaxint = canonicalize_dtype(int)
jaxfloat = canonicalize_dtype(float)

integers_strategy = st.integers(
    min_value=np.iinfo(jaxint).min, max_value=np.iinfo(jaxint).max
)
floats_strategy = st.floats(
    min_value=np.finfo(jaxfloat).min, max_value=np.finfo(jaxfloat).max
)


@given(
    value=integers_strategy
    | floats_strategy
    | st.lists(integers_strategy)
    | st.tuples(integers_strategy)
    | st.lists(floats_strategy)
    | st.tuples(floats_strategy)
    # | st.lists(st.lists(integers_strategy))  # TODO: enable nested lists
    # | st.lists(st.lists(floats_strategy))
    | np_arrays(
        dtype=np.float32,
        shape=np_array_shapes(),
        elements={"allow_nan": False, "allow_infinity": False},
    )
    | xps.arrays(
        dtype=np.float32,
        shape=xps.array_shapes(),
        elements={"allow_nan": False, "allow_infinity": False},
    )
)
@example(value=0)  # int
@example(value=1.0)  # int
@example(value=[1])  # list[int]
@example(value=(1,))  # tuple[int, ...]
@example(value=[1.0])  # list[float]
@example(value=(1.0,))  # list[float]
@example(value=[[1]])  # list[list[int]]
@example(value=[[1.0]])  # list[list[int]]
def test_properties(value):
    """Test the properties of Quantity."""
    q = u.Quantity(value, "m")
    expected = jnp.asarray(value)

    # Test the value
    assert np.array_equal(q.value, expected)

    # Test the shape
    assert q.shape == expected.shape

    # Test materialise
    with pytest.raises(RuntimeError):
        q.materialise()

    # Test aval
    assert q.aval() == jax.core.get_aval(expected)

    # Test enable_materialise
    assert np.array_equal(q.enable_materialise().value, q.value)


def test_parametric():
    """Test the parametric strategy."""
    # Inferred
    q = u.Quantity(1, "m")
    (dims,) = q._type_parameter
    assert dims == u.dimension("length")

    # Explicit
    q = u.Quantity["length"](1, "m")
    (dims,) = q._type_parameter
    assert dims == u.dimension("length")

    q = u.Quantity["length"](jnp.ones((1, 2)), "m")
    (dims,) = q._type_parameter
    assert dims == u.dimension("length")

    # type-checks
    with pytest.raises(ValueError, match="Physical type mismatch."):
        u.Quantity["time"](1, "m")


@pytest.mark.parametrize("unit", [apyu.m, "meter"])
def test_unit(unit):
    """Test the unit."""
    assert u.Quantity(1, unit).unit == unit


def test_array_namespace():
    """Test the array namespace."""
    assert u.Quantity(1, "m").__array_namespace__() is jnp


def test_uconvert():
    """Test the ``u.Quantity.uconvert`` method."""
    q = u.Quantity(1, "m")
    assert jnp.equal(q.uconvert("km"), u.Quantity(0.001, "km"))


def test_ustrip():
    """Test the ``u.Quantity.ustrip`` method."""
    q = u.Quantity(1, "m")
    assert q.ustrip("km") == u.Quantity(0.001, "km").value


@pytest.mark.skip("TODO")
def test_getitem():
    """Test the ``u.Quantity.__getitem__`` method."""
    raise NotImplementedError


def test_len():
    """Test the ``len(Quantity)`` method."""
    # Length 3
    q = u.Quantity([1, 2, 3], "m")
    assert len(q) == 3

    # Scalar
    q = u.Quantity(1, "m")
    assert len(q) == 0


@pytest.mark.skip("TODO")
def test_add():
    """Test the ``Quantity.__add__`` method."""
    raise NotImplementedError


@pytest.mark.skip("TODO")
def test_radd():
    """Test the ``Quantity.__radd__`` method."""
    raise NotImplementedError


@pytest.mark.skip("TODO")
def test_sub():
    """Test the ``Quantity.__sub__`` method."""
    raise NotImplementedError


@pytest.mark.skip("TODO")
def test_rsub():
    """Test the ``Quantity.__rsub__`` method."""
    raise NotImplementedError


@pytest.mark.skip("TODO")
def test_mul():
    """Test the ``Quantity.__mul__`` method."""
    raise NotImplementedError


@pytest.mark.skip("TODO")
def test_rmul():
    """Test the ``Quantity.__rmul__`` method."""
    raise NotImplementedError


@pytest.mark.skip("TODO")
def test_matmul():
    """Test the ``Quantity.__matmul__`` method."""
    raise NotImplementedError


@pytest.mark.skip("TODO")
def test_rmatmul():
    """Test the ``Quantity.__rmatmul__`` method."""
    raise NotImplementedError


@pytest.mark.skip("TODO")
def test_pow():
    """Test the ``Quantity.__pow__`` method."""
    raise NotImplementedError


@pytest.mark.skip("TODO")
def test_rpow():
    """Test the ``Quantity.__rpow__`` method."""
    raise NotImplementedError


@pytest.mark.skip("TODO")
def test_truediv():
    """Test the ``Quantity.__truediv__`` method."""
    raise NotImplementedError


@pytest.mark.skip("TODO")
def test_rtruediv():
    """Test the ``Quantity.__rtruediv__`` method."""
    raise NotImplementedError


@pytest.mark.skip("TODO")
def test_and():
    """Test the ``Quantity.__and__`` method."""
    raise NotImplementedError


def test_gt():
    """Test the ``Quantity.__gt__`` method."""
    # Test with a scalar
    q = u.Quantity(1, "m")
    assert q > u.Quantity(0, "m")
    assert not q > u.Quantity(1, "m")
    assert not q > u.Quantity(2, "m")

    # Test with an array
    q = u.Quantity([1, 2, 3], "m")
    assert np.array_equal(q > u.Quantity(0, "m"), [True, True, True])
    assert np.array_equal(q > u.Quantity(1, "m"), [False, True, True])

    # Test with incompatible units
    # TODO: better equinox exception matching
    with pytest.raises(Exception):  # noqa: B017, PT011
        _ = q > u.Quantity(0, "s")

    # Test special case w/out units
    assert u.Quantity(1, "m") > 0
    assert np.array_equal(u.Quantity([-1, 0, 1], "m") > 0, [False, False, True])


def test_ge():
    """Test the ``Quantity.__ge__`` method."""
    # Test with a scalar
    q = u.Quantity(1, "m")
    assert q >= u.Quantity(0, "m")
    assert q >= u.Quantity(1, "m")
    assert not q >= u.Quantity(2, "m")

    # Test with an array
    q = u.Quantity([1, 2, 3], "m")
    assert np.array_equal(q >= u.Quantity(0, "m"), [True, True, True])
    assert np.array_equal(q >= u.Quantity(1, "m"), [True, True, True])
    assert np.array_equal(q >= u.Quantity(2, "m"), [False, True, True])

    # Test with incompatible units
    # TODO: better equinox exception matching
    with pytest.raises(Exception):  # noqa: B017, PT011
        _ = q >= u.Quantity(0, "s")

    # Test special case w/out units
    assert u.Quantity(1, "m") >= 0
    assert np.array_equal(u.Quantity([-1, 0, 1], "m") >= 0, [False, True, True])


def test_lt():
    """Test the ``Quantity.__lt__`` method."""
    # Test with a scalar
    q = u.Quantity(1, "m")
    assert not q < u.Quantity(0, "m")
    assert not q < u.Quantity(1, "m")
    assert q < u.Quantity(2, "m")

    # Test with an array
    q = u.Quantity([1, 2, 3], "m")
    assert np.array_equal(q < u.Quantity(0, "m"), [False, False, False])
    assert np.array_equal(q < u.Quantity(1, "m"), [False, False, False])
    assert np.array_equal(q < u.Quantity(2, "m"), [True, False, False])

    # Test with incompatible units
    # TODO: better equinox exception matching
    with pytest.raises(Exception):  # noqa: B017, PT011
        _ = q < u.Quantity(0, "s")

    # Test special case w/out units
    assert u.Quantity(-1, "m") < 0
    assert np.array_equal(u.Quantity([-1, 0, 1], "m") < 0, [True, False, False])


def test_le():
    """Test the ``u.Quantity.__le__`` method."""
    # Test with a scalar
    q = u.Quantity(1, "m")
    assert not q <= u.Quantity(0, "m")
    assert q <= u.Quantity(1, "m")
    assert q <= u.Quantity(2, "m")

    # Test with an array
    q = u.Quantity([1, 2, 3], "m")
    assert np.array_equal(q <= u.Quantity(0, "m"), [False, False, False])
    assert np.array_equal(q <= u.Quantity(1, "m"), [True, False, False])
    assert np.array_equal(q <= u.Quantity(2, "m"), [True, True, False])

    # Test with incompatible units
    # TODO: better equinox exception matching
    with pytest.raises(Exception):  # noqa: B017, PT011
        _ = q <= u.Quantity(0, "s")

    # Test special case w/out units
    assert u.Quantity(-1, "m") <= 0
    assert np.array_equal(u.Quantity([-1, 0, 1], "m") <= 0, [True, True, False])


def test_eq():
    """Test the ``Quantity.__eq__`` method."""
    # Test with a scalar
    q = u.Quantity(1, "m")
    assert not q == u.Quantity(0, "m")  # noqa: SIM201
    assert q == u.Quantity(1, "m")
    assert not q == u.Quantity(2, "m")  # noqa: SIM201

    # Test with an array
    q = u.Quantity([1, 2, 3], "m")
    assert np.array_equal(q == u.Quantity(0, "m"), [False, False, False])
    assert np.array_equal(q == u.Quantity(1, "m"), [True, False, False])
    assert np.array_equal(q == u.Quantity(2, "m"), [False, True, False])

    # Test with incompatible units
    with pytest.raises(Exception):  # noqa: B017, PT011
        _ = jnp.equal(q, u.Quantity(0, "s"))

    # Test special case w/out units
    assert u.Quantity(0, "m") == 0
    assert np.array_equal(u.Quantity([-1, 0, 1], "m") == 0, [False, True, False])


def test_ne():
    """Test the ``Quantity.__ne__`` method."""
    # Test with a scalar
    q = u.Quantity(1, "m")
    assert q != u.Quantity(0, "m")
    assert q == u.Quantity(1, "m")
    assert q != u.Quantity(2, "m")

    # Test with an array
    q = u.Quantity([1, 2, 3], "m")
    assert np.array_equal(q != u.Quantity(0, "m"), [True, True, True])
    assert np.array_equal(q != u.Quantity(1, "m"), [False, True, True])
    assert np.array_equal(q != u.Quantity(2, "m"), [True, False, True])

    # Test with incompatible units
    # TODO: better equinox exception matching
    with pytest.raises(Exception):  # noqa: B017, PT011
        _ = jnp.not_equal(q, u.Quantity(0, "s"))
    with pytest.raises(Exception):  # noqa: B017, PT011
        _ = jnp.not_equal(q, u.Quantity(4, "s"))

    # Test special case w/out units
    assert u.Quantity(1, "m") != 0
    assert np.array_equal(u.Quantity([-1, 0, 1], "m") != 0, [True, False, True])


def test_neg():
    """Test the ``Quantity.__neg__`` method."""
    # Test with a scalar
    q = u.Quantity(1, "m")
    assert -q == u.Quantity(-1, "m")

    # Test with an array
    q = u.Quantity([1, 2, 3], "m")
    assert np.array_equal(-q.value, [-1, -2, -3])
    assert (-q).unit == u.unit("m")


def test_flatten():
    """Test the ``Quantity.flatten`` method."""
    # Test with a scalar
    q = u.Quantity(1, "m")
    assert q.flatten() == u.Quantity(1, "m")

    # Test with an array
    q = u.Quantity([[1, 2, 3], [4, 5, 6]], "m")
    assert np.array_equal(q.flatten().value, [1, 2, 3, 4, 5, 6])
    assert q.flatten().unit == u.unit("m")


def test_reshape():
    """Test the ``Quantity.reshape`` method."""
    # Test with a scalar
    q = u.Quantity(1, "m")
    assert q.reshape(1, 1) == u.Quantity(1, "m")

    # Test with an array
    q = u.Quantity([1, 2, 3, 4, 5, 6], "m")
    assert np.array_equal(q.reshape(2, 3).value, [[1, 2, 3], [4, 5, 6]])
    assert q.reshape(2, 3).unit == u.unit("m")


def test_hypot():
    """Test the ``jnp.hypot`` method."""
    q1 = u.Quantity(3, "m")
    q2 = u.Quantity(4, "m")
    assert jnp.hypot(q1, q2) == u.Quantity(5, "m")

    q1 = u.Quantity([1, 2, 3], "m")
    q2 = u.Quantity([4, 5, 6], "m")
    assert all(jnp.hypot(q1, q2) == u.Quantity([4.1231055, 5.3851647, 6.7082043], "m"))


def test_mod():
    """Test taking the modulus."""
    q = u.Quantity(480.0, "deg")

    with pytest.raises(AttributeError):
        _ = q % 2

    with pytest.raises(apyu.UnitConversionError):
        _ = q % u.Quantity(2, "m")

    got = q % u.Quantity(360, "deg")
    expect = u.Quantity(120, "deg")
    assert got == expect


# --------------------------------------------------------------


def test_at():
    """Test the ``Quantity.at`` method."""
    x = jnp.arange(10, dtype=float)
    q = u.Quantity(x, "km")

    # Get
    # TODO: test fill_value
    assert q.at[1].get() == u.Quantity(1.0, "km")
    assert np.array_equal(q.at[:3].get(), u.Quantity([0.0, 1, 2], "km"))

    # Set
    q2 = q.at[1].set(u.Quantity(1.2, "km"))
    assert q2[1] == u.Quantity(1.2, "km")
    assert q[1] == u.Quantity(1.0, "km")  # original is unchanged

    q2 = q.at[:3].set(u.Quantity([1.2, 2.3, 3.4], "km"))
    assert np.array_equal(q2[:3], u.Quantity([1.2, 2.3, 3.4], "km"))
    assert np.array_equal(q[:3], u.Quantity([0.0, 1, 2], "km"))  # original is unchanged

    # Apply
    with pytest.raises(NotImplementedError):
        q.at[1].apply(lambda x: x + 1)

    # Add
    q2 = q.at[1].add(u.Quantity(1.2, "km"))
    assert q2[1] == u.Quantity(2.2, "km")
    assert q[1] == u.Quantity(1.0, "km")  # original is unchanged

    # Multiply
    q2 = q.at[1].mul(2)
    assert q2[1] == u.Quantity(2.0, "km")
    assert q[1] == u.Quantity(1.0, "km")  # original is unchanged

    with pytest.raises((RuntimeError, TypeCheckError)):
        q.at[1].mul(u.Quantity(2, "m"))

    # Divide
    q2 = q.at[1].divide(2)
    assert q2[1] == u.Quantity(0.5, "km")
    assert q[1] == u.Quantity(1.0, "km")  # original is unchanged

    with pytest.raises((RuntimeError, TypeCheckError)):
        q.at[1].divide(u.Quantity(2, "m"))

    # Power
    with pytest.raises(NotImplementedError):
        q.at[1].power(2)

    # Min
    q2 = q.at[1].min(u.Quantity(0.5, "km"))
    assert q2[1] == u.Quantity(0.5, "km")
    assert q[1] == u.Quantity(1.0, "km")  # original is unchanged

    q2 = q.at[1].min(u.Quantity(1.5, "km"))
    assert q2[1] == u.Quantity(1.0, "km")
    assert q[1] == u.Quantity(1.0, "km")  # original is unchanged

    # Max
    q2 = q.at[1].max(u.Quantity(1.5, "km"))
    assert q2[1] == u.Quantity(1.5, "km")
    assert q[1] == u.Quantity(1.0, "km")  # original is unchanged

    q2 = q.at[1].max(u.Quantity(0.5, "km"))
    assert q2[1] == u.Quantity(1.0, "km")
    assert q[1] == u.Quantity(1.0, "km")  # original is unchanged


# ---------------------------


@parametric
class NewQuantity(AbstractParametricQuantity):
    """Quantity with a flag."""

    value: Array = eqx.field(converter=jnp.asarray)
    unit: str = eqx.field(converter=u.unit)
    flag: bool = eqx.field(static=True, kw_only=True)


def test_parametric_pickle_dumps_with_kw_fields():
    x = NewQuantity([1, 2, 3], "m", flag=True)
    assert isinstance(pickle.dumps(x), bytes)


# ===============================================================
# Astropy


def test_from_astropy():
    """Test the ``Quantity.from_(AstropyQuantity)`` method."""
    apyq = apyu.Quantity(1, "m")
    q = u.Quantity.from_(apyq)
    assert isinstance(q, u.Quantity)
    assert np.equal(q.value, apyq.value)
    assert q.unit == apyq.unit


def test_convert_to_astropy():
    """Test the ``convert(Quantity, AstropyQuantity)`` method."""
    q = u.Quantity(1, "m")
    apyq = convert(q, apyu.Quantity)
    assert isinstance(apyq, apyu.Quantity)
    assert apyq == apyu.Quantity(1, "m")


##############################################################################


def test_is_unit_convertible():
    """Test `unxt.is_unit_convertible`."""
    # Unit
    assert u.is_unit_convertible(apyu.km, apyu.kpc) is True

    # unit is a str
    assert u.is_unit_convertible("km", "m") is True

    # Bad unit
    assert u.is_unit_convertible(apyu.s, apyu.m) is False

    # Quantity
    assert u.is_unit_convertible(apyu.kpc, u.Quantity(1, "km")) is True

    # unit is a str
    assert u.is_unit_convertible("km", u.Quantity(1, "km")) is True

    # Bad quantity
    assert u.is_unit_convertible(apyu.m, u.Quantity(1, "s")) is False



"""Units objects in JAX.

Copyright (c) 2023 Galactic Dynamics. All rights reserved.
"""

__all__: list[str] = []

from typing import Any

import astropy.units as apyu
from plum import dispatch

from .api import AbstractDimension

# ===================================================================
# Construct the dimensions


@dispatch
def dimension(obj: AbstractDimension, /) -> AbstractDimension:
    """Construct dimension from a dimension object.

    Examples
    --------
    >>> import unxt as u
    >>> import astropy.units as apyu

    >>> length = apyu.get_physical_type("length")
    >>> length
    PhysicalType('length')

    >>> u.dimension(length) is length
    True

    """
    return obj


@dispatch
def dimension(obj: str, /) -> AbstractDimension:
    """Construct dimension from a string.

    Examples
    --------
    >>> from unxt.dims import dimension
    >>> dimension("length")
    PhysicalType('length')

    """
    return apyu.get_physical_type(obj)


# ===================================================================
# Get the dimension


@dispatch
def dimension_of(obj: Any, /) -> None:
    """Most objects have no dimension.

    Examples
    --------
    >>> from unxt.dims import dimension_of

    >>> print(dimension_of(1))
    None

    >>> print(dimension_of("length"))
    None

    """
    return None  # noqa: RET501


@dispatch
def dimension_of(obj: AbstractDimension, /) -> AbstractDimension:
    """Return the dimension of the given units.

    Examples
    --------
    >>> from unxt.dims import dimension, dimension_of

    >>> dimension_of(dimension("length"))
    PhysicalType('length')

    """
    return obj



"""Benchmark tests for quaxified jax."""

import equinox as eqx
import jax
import pytest
from jaxlib.xla_extension import PjitFunction

import unxt as u


@pytest.fixture
def func_dimension() -> PjitFunction:
    return eqx.filter_jit(u.dimension)


@pytest.fixture
def func_dimension_of() -> PjitFunction:
    return eqx.filter_jit(u.dimension_of)


#####################################################################
# `dimension`

args = [(u.dimension("length"),), ("length",)]


@pytest.mark.parametrize("args", args, ids=str)
@pytest.mark.benchmark(group="dimensions", warmup=False, max_time=1.0)
def test_dimension(args):
    """Test calling `unxt.dimension`."""
    _ = u.dimension(*args)


@pytest.mark.parametrize("args", args, ids=str)
@pytest.mark.benchmark(group="dimensions", warmup=False, max_time=1.0)
def test_dimension_jit_compile(func_dimension, args):
    """Test the speed of jitting."""
    _ = func_dimension.lower(*args).compile()


@pytest.mark.parametrize("args", args, ids=str)
@pytest.mark.benchmark(group="dimensions", warmup=True, max_time=1.0)
def test_dimension_execute(func_dimension, args):
    """Test the speed of calling the function."""
    _ = jax.block_until_ready(func_dimension(*args))


#####################################################################
# `dimension_of`


args = [
    (u.dimension("length"),),  # -> Dimension('length')
    (u.unit("m"),),  # -> Dimension('length')
    (u.Quantity(1, "m"),),  # -> Dimension('length')
    (2,),  # -> None
]


@pytest.mark.parametrize("args", args, ids=str)
@pytest.mark.benchmark(group="dimensions", warmup=False, max_time=1.0)
def test_dimension_of(args):
    """Test calling `unxt.dimension_of`."""
    _ = u.dimension_of(*args)


@pytest.mark.parametrize("args", args, ids=str)
@pytest.mark.benchmark(group="dimensions", warmup=False, max_time=1.0)
def test_dimension_of_jit_compile(func_dimension_of, args):
    """Test the speed of jitting."""
    _ = func_dimension_of.lower(*args).compile()


@pytest.mark.parametrize("args", args, ids=str)
@pytest.mark.benchmark(group="dimensions", warmup=True, max_time=1.0)
def test_dimension_of_execute(func_dimension_of, args):
    """Test the speed of calling the function."""
    _ = jax.block_until_ready(func_dimension_of(*args))



"""Compatibility for Quantity."""

__all__: list[str] = []

from plum import conversion_method

from .api import ustrip
from .base import AbstractQuantity
from .quantity import Quantity
from .unchecked import BareQuantity
from unxt.units import unit_of


@conversion_method(type_from=AbstractQuantity, type_to=BareQuantity)  # type: ignore[arg-type]
def quantity_to_unchecked(q: AbstractQuantity, /) -> BareQuantity:
    """Convert any quantity to an unchecked quantity.

    Examples
    --------
    >>> from plum import convert
    >>> from unxt.quantity import Quantity, BareQuantity

    >>> q = Quantity(1, "m")
    >>> convert(q, BareQuantity)
    BareQuantity(Array(1, dtype=int32, weak_type=True), unit='m')

    The self-conversion doesn't copy the object:

    >>> q = BareQuantity(1, "m")
    >>> convert(q, BareQuantity) is q
    True

    """
    if isinstance(q, BareQuantity):
        return q
    unit = unit_of(q)
    return BareQuantity(ustrip(unit, q), unit)


@conversion_method(type_from=AbstractQuantity, type_to=Quantity)  # type: ignore[arg-type]
def quantity_to_checked(q: AbstractQuantity, /) -> Quantity:
    """Convert any quantity to a checked quantity.

    Examples
    --------
    >>> from plum import convert
    >>> from unxt.quantity import Quantity, BareQuantity

    >>> q = BareQuantity(1, "m")
    >>> q
    BareQuantity(Array(1, dtype=int32, ...), unit='m')

    >>> convert(q, Quantity)
    Quantity['length'](Array(1, dtype=int32, ...), unit='m')

    The self-conversion doesn't copy the object:

    >>> q = Quantity(1, "m")
    >>> convert(q, Quantity) is q
    True

    """
    if isinstance(q, Quantity):
        return q
    u = unit_of(q)
    return Quantity(ustrip(u, q), u)



"""Tools for representing systems of units using ``astropy.units``."""

__all__ = ["unitsystem", "unitsystem_of"]


from collections.abc import Sequence
from dataclasses import field, make_dataclass
from typing import Annotated, Any

import equinox as eqx
import numpy as np
from astropy.constants import G as const_G  # noqa: N811, pylint: disable=E0611
from astropy.units import UnitBase as AstropyUnitBase
from plum import dispatch

from . import builtin_dimensions as ud
from .base import UNITSYSTEMS_REGISTRY, AbstractUnitSystem
from .builtin import DimensionlessUnitSystem
from .flags import AbstractUSysFlag, DynamicalSimUSysFlag, StandardUSysFlag
from .realizations import NAMED_UNIT_SYSTEMS, dimensionless
from .utils import get_dimension_name
from unxt.dims import dimension_of
from unxt.units import unit

# ===================================================================
# `unitsystem` function


@dispatch
def unitsystem(usys: AbstractUnitSystem, /) -> AbstractUnitSystem:
    """Convert a UnitSystem to a UnitSystem.

    Examples
    --------
    >>> from unxt.unitsystems import unitsystem
    >>> usys = unitsystem("kpc", "Myr", "Msun", "radian")
    >>> usys
    unitsystem(kpc, Myr, solMass, rad)

    >>> unitsystem(usys) is usys
    True

    """
    return usys


@dispatch
def unitsystem(seq: Sequence[Any], /) -> AbstractUnitSystem:
    """Convert a UnitSystem or tuple of arguments to a UnitSystem.

    Examples
    --------
    >>> import unxt as u

    >>> u.unitsystem(())
    DimensionlessUnitSystem()

    >>> u.unitsystem(("kpc", "Myr", "Msun", "radian"))
    unitsystem(kpc, Myr, solMass, rad)

    >>> u.unitsystem(["kpc", "Myr", "Msun", "radian"])
    unitsystem(kpc, Myr, solMass, rad)

    """
    return unitsystem(*seq) if len(seq) > 0 else dimensionless


@dispatch
def unitsystem(_: None, /) -> DimensionlessUnitSystem:
    """Dimensionless unit system from None.

    Examples
    --------
    >>> from unxt.unitsystems import unitsystem
    >>> unitsystem(None)
    DimensionlessUnitSystem()

    """
    return dimensionless


@dispatch
def unitsystem(*args: Any) -> AbstractUnitSystem:
    """Convert a set of arguments to a UnitSystem.

    Examples
    --------
    >>> from unxt.unitsystems import unitsystem

    >>> unitsystem("kpc", "Myr", "Msun", "radian")
    unitsystem(kpc, Myr, solMass, rad)

    """
    # Convert everything to a unit
    args = tuple(map(unit, args))

    # Check that the units all have different dimensions
    dims = tuple(map(dimension_of, args))
    dims = eqx.error_if(
        dims,
        len(set(dims)) < len(dims),
        "some dimensions are repeated",
    )

    # Return if the unit system is already registered
    if dims in UNITSYSTEMS_REGISTRY:
        return UNITSYSTEMS_REGISTRY[dims](*args)

    # Otherwise, create a new unit system
    # dimension names of all the units
    du = {get_dimension_name(x).replace(" ", "_"): dimension_of(x) for x in args}
    # name: physical types
    cls_name = "".join(k.title().replace("_", "") for k in du) + "UnitSystem"
    # fields: name, unit
    fields = [
        (
            k,
            Annotated[AstropyUnitBase, v],
            field(init=True, repr=True, hash=True, compare=True),  # pylint: disable=invalid-field-call
        )
        for k, v in du.items()
    ]

    def _reduce_(self: AbstractUnitSystem) -> tuple:
        return (_call_unitsystem, self.base_units, None, None, None, None)

    # Make and register the dataclass class
    unitsystem_cls: type[AbstractUnitSystem] = make_dataclass(
        cls_name,
        fields,
        bases=(AbstractUnitSystem,),
        namespace={"__reduce__": _reduce_},
        frozen=True,
        slots=True,
        eq=True,
        repr=True,
        init=True,
    )

    # Make the dataclass instance
    return unitsystem_cls(*args)


@dispatch
def unitsystem(name: str, /) -> AbstractUnitSystem:
    """Return unit system from name.

    Examples
    --------
    >>> from unxt.unitsystems import unitsystem
    >>> unitsystem("galactic")
    unitsystem(kpc, Myr, solMass, rad)

    >>> unitsystem("solarsystem")
    unitsystem(AU, yr, solMass, rad)

    >>> unitsystem("dimensionless")
    DimensionlessUnitSystem()

    """
    return NAMED_UNIT_SYSTEMS[name]


@dispatch
def unitsystem(usys: AbstractUnitSystem, *args: Any) -> AbstractUnitSystem:
    """Create a unit system from an existing unit system and additional units.

    Examples
    --------
    We can add a new unit definition to an existing unit system:

    >>> from unxt.unitsystems import unitsystem
    >>> usys = unitsystem("galactic")
    >>> unitsystem(usys, "km/s")
    LengthTimeMassAngleSpeedUnitSystem(length=Unit("kpc"), time=Unit("Myr"), mass=Unit("solMass"), angle=Unit("rad"), speed=Unit("km / s"))

    We can also override the base unit of an existing unit system:

    >>> new_usys = unitsystem(usys, "pc")
    >>> new_usys
    TimeMassAngleLengthUnitSystem(time=Unit("Myr"), mass=Unit("solMass"), angle=Unit("rad"), length=Unit("pc"))

    """  # noqa: E501
    # TODO: not need this hack for single-string inputs
    # TODO: process new units without making a whole unit system
    if len(args) == 1 and isinstance(args[0], str):
        new_usys = unitsystem(unit(args[0]))
    else:
        new_usys = unitsystem(*args)

    current_units = [
        u for u in usys.base_units if dimension_of(u) not in new_usys.base_dimensions
    ]
    return unitsystem(*current_units, *args)


@dispatch
def unitsystem(flag: type[AbstractUSysFlag], *_: Any) -> AbstractUnitSystem:
    """Raise an exception since the flag is abstract."""
    msg = "Do not use the AbstractUSysFlag directly, only use subclasses."
    raise TypeError(msg)


@dispatch
def unitsystem(flag: type[StandardUSysFlag], *args: Any) -> AbstractUnitSystem:
    """Create a standard unit system using the inputted units.

    Examples
    --------
    >>> from unxt import unitsystem, unitsystems
    >>> unitsystem(unitsystems.StandardUSysFlag, "kpc", "Myr", "Msun")
    LengthTimeMassUnitSystem(length=Unit("kpc"), time=Unit("Myr"), mass=Unit("solMass"))

    """
    return unitsystem(*args)


@dispatch
def unitsystem(
    flag: type[DynamicalSimUSysFlag],
    *args: Any,
    G: float | int = 1.0,  # noqa: N803
) -> AbstractUnitSystem:
    """Make a dynamical unit system.

    Examples
    --------
    >>> from unxt.unitsystems import unitsystem, DynamicalSimUSysFlag

    >>> unitsystem(DynamicalSimUSysFlag, "m", "kg")
    LengthMassTimeUnitSystem(length=Unit("m"), mass=Unit("kg"), time=Unit("122404 s"))

    """
    tmp = unitsystem(*args)

    # Use G for computing the missing units below:
    G = G * const_G

    added = ()
    if ud.length in tmp.base_dimensions and ud.mass in tmp.base_dimensions:
        time = 1 / np.sqrt(G * tmp["mass"] / tmp["length"] ** 3)
        added = (time,)
    elif ud.length in tmp.base_dimensions and ud.time in tmp.base_dimensions:
        mass = 1 / G * tmp["length"] ** 3 / tmp["time"] ** 2
        added = (mass,)
    elif ud.length in tmp.base_dimensions and ud.speed in tmp.base_dimensions:
        time = tmp["length"] / tmp["velocity"]
        mass = tmp["velocity"] ** 2 / G * tmp["length"]
        added = (time, mass)
    elif ud.mass in tmp.base_dimensions and ud.time in tmp.base_dimensions:
        length = np.cbrt(G * tmp["mass"] * tmp["time"] ** 2)
        added = (length,)
    elif ud.mass in tmp.base_dimensions and ud.speed in tmp.base_dimensions:
        length = G * tmp["mass"] / tmp["velocity"] ** 2
        time = length / tmp["velocity"]
        added = (length, time)
    elif ud.time in tmp.base_dimensions and ud.speed in tmp.base_dimensions:
        mass = 1 / G * tmp["velocity"] ** 3 * tmp["time"]
        length = G * mass / tmp["velocity"] ** 2
        added = (mass, length)

    return unitsystem(*tmp, *added)


# ----


def _call_unitsystem(*args: Any) -> AbstractUnitSystem:
    return unitsystem(*args)


# ===================================================================
# `unitsystem_of` function


@dispatch.abstract
def unitsystem_of(obj: Any, /) -> AbstractUnitSystem:
    """Return the unit system of the object."""


@dispatch
def unitsystem_of(obj: Any, /) -> DimensionlessUnitSystem:
    """Return the unit system of the object.

    Examples
    --------
    >>> from unxt.unitsystems import unitsystem_of

    >>> unitsystem_of(1)
    DimensionlessUnitSystem()

    """
    return dimensionless


@dispatch
def unitsystem_of(obj: AbstractUnitSystem, /) -> AbstractUnitSystem:
    """Return the unit system from the unit system.

    Examples
    --------
    >>> from unxt.unitsystems import galactic, unitsystem_of

    >>> unitsystem_of(galactic) is galactic
    True

    """
    return obj
